from typing import List, AnyStr

from bolinette.routing import Route, Method, AccessType, resources
from bolinette.routing.defaults import Defaults


class Namespace:
    def __init__(self, base_url, service):
        self.base_url = base_url
        self.service = service
        self.model = service.name
        self.route = NamespaceRoute(self)

    @property
    def defaults(self):
        return Defaults(self)


class NamespaceExcepts:
    def __init__(self, model, key='default', *, patch=False):
        self.model = model
        self.key = key
        self.patch = patch


class NamespaceReturns:
    def __init__(self, model, key='default', *, as_list=False, skip_none=False):
        self.model = model
        self.key = key
        self.as_list = as_list
        self.skip_none = skip_none


class NamespaceRoute:
    def __init__(self, namespace):
        self.namespace = namespace

    def __call__(self, path, *, method: Method,
                 access: AccessType = None, expects: NamespaceExcepts = None,
                 returns: NamespaceReturns = None, roles: List[AnyStr] = None):
        def inner(func):
            route = Route(func=func, base_url=self.namespace.base_url, path=path, method=method,
                          access=access, expects=expects, returns=returns, roles=roles)
            resources.register(route)
            return func
        return inner

    def expects(self, model, key='default', *, patch=False):
        return NamespaceExcepts(model, key, patch=patch)

    def returns(self, model, key='default', *, as_list=False, skip_none=False):
        return NamespaceReturns(model, key, as_list=as_list, skip_none=skip_none)



from bolinette import db
from bolinette.commands import command


@command('init_db')
async def init_db(*, run_seeders=False, **_):
    await db.engine.drop_all()
    await db.engine.create_all()
    if run_seeders:
        await db.engine.run_seeders()



from bolinette import mapping, db


class File(db.defs.model):
    __tablename__ = 'file'

    id = db.defs.column(db.types.integer, primary_key=True)
    key = db.defs.column(db.types.string, nullable=False)
    name = db.defs.column(db.types.string, nullable=False)
    mime = db.defs.column(db.types.string, nullable=False)

    @staticmethod
    def responses():
        yield [
            mapping.Field(db.types.string, key='key'),
            mapping.Field(db.types.string, key='name'),
            mapping.Field(db.types.string, key='mime')
        ]
        yield 'minimal', [
            mapping.Field(db.types.string, key='key')
        ]


mapping.register(File)



from bolinette.models import Role
from bolinette.services import BaseService


class RoleService(BaseService):
    def __init__(self):
        super().__init__(Role)

    async def get_by_name(self, name):
        return await self.get_first_by('name', name)


role_service = RoleService()



from bolinette import mapping, db
from bolinette.models import User
from sqlalchemy.ext.declarative.base import declared_attr


class Historized:
    created_on = db.defs.column(db.types.date)
    updated_on = db.defs.column(db.types.date)

    @declared_attr
    def created_by_id(self):
        return db.defs.column(db.types.integer, db.types.foreign_key('user', 'id'))

    @declared_attr
    def created_by(self):
        return db.defs.relationship(User, foreign_keys=self.created_by_id, lazy=False)

    @declared_attr
    def updated_by_id(self):
        return db.defs.column(db.types.integer, db.types.foreign_key('user', 'id'))

    @declared_attr
    def updated_by(self):
        return db.defs.relationship(User, foreign_keys=self.updated_by_id, lazy=False)

    @staticmethod
    def base_response():
        return [
            mapping.Field(db.types.date, key='created_on'),
            mapping.Field(db.types.date, key='updated_on'),
            mapping.Definition('created_by', 'user'),
            mapping.Definition('updated_by', 'user'),
        ]



from enum import unique, Enum, auto


@unique
class Method(Enum):
    GET = auto()
    POST = auto()
    PUT = auto()
    PATCH = auto()
    DELETE = auto()

    @property
    def http_verb(self):
        if self.value == Method.GET.value:
            return 'GET'
        elif self.value == Method.POST.value:
            return 'POST'
        elif self.value == Method.PUT.value:
            return 'PUT'
        elif self.value == Method.PATCH.value:
            return 'PATCH'
        elif self.value == Method.DELETE.value:
            return 'DELETE'



import asyncio
import inspect

from aiohttp import web as aio_web

from bolinette import env, jwt, bcrypt, db
from bolinette.commands import commands
from bolinette.routing import web, resources


class Bolinette:
    def __init__(self, *, profile=None, overrides=None):
        env.init_app(profile=profile, overrides=overrides)
        db.engine.init_app()
        jwt.init_app()
        bcrypt.init_app()
        web.init_app()
        resources.init_app()

    def run(self):
        aio_web.run_app(web.app)

    @property
    def app(self):
        return web.app

    def run_command(self, name, **kwargs):
        if name in commands.commands:
            func = commands.commands[name]
            if inspect.isfunction(func):
                if inspect.iscoroutinefunction(func):
                    loop = asyncio.get_event_loop()
                    loop.run_until_complete(func(**kwargs))
                else:
                    func(**kwargs)



from types import SimpleNamespace

import aiohttp_cors

from bolinette.routing import web


class Resources:
    def __init__(self):
        self.routes = []
        self.res = None
        self.cors_default = {
            "*": aiohttp_cors.ResourceOptions(
                allow_credentials=True,
                expose_headers="*",
                allow_headers="*",
            )
        }

    def register(self, route):
        self.routes.append(route)

    def init_app(self):
        self.res = {}
        for route in self.routes:
            path = f'/api{route.base_url}{route.path}'
            method = route.method
            handler = route.handler
            if path not in self.res:
                self.res[path] = SimpleNamespace()
                self.res[path].res = web.cors.add(web.app.router.add_resource(path))
                self.res[path].routes = {}
            self.res[path].routes[method.http_verb] = web.cors.add(
                self.res[path].res.add_route(method.http_verb, handler), self.cors_default)


resources = Resources()



import os


def read_version(path):
    try:
        with open(join(path, '.version')) as f:
            for line in f:
                return line.strip().replace('\n', '')
    except FileNotFoundError:
        return None


def cwd():
    return os.getcwd()


def dirname(path):
    return os.path.dirname(os.path.realpath(path))


def join(*args):
    return os.path.join(*args)


def mkdir(path):
    os.makedirs(path)


def exists(path):
    return os.path.exists(path)


def delete(path):
    os.unlink(path)



import string
from datetime import datetime
from random import random

from bolinette import response, jwt, Cookie, mail, env
from bolinette.exceptions import EntityNotFoundError, BadRequestError
from bolinette.routing import Namespace, Method, AccessType
from bolinette.services import user_service, role_service

ns = Namespace('/user', user_service)


def _create_tokens(resp, user, *, set_access, set_refresh, fresh):
    now = datetime.utcnow()
    if set_access:
        access_token = jwt.create_access_token(now, user.username, fresh=fresh)
        resp.cookies.append(Cookie('access_token', access_token,
                                   expires=jwt.access_token_expires(now), path='/api'))
    if set_refresh:
        refresh_token = jwt.create_refresh_token(now, user.username)
        resp.cookies.append(Cookie('refresh_token', refresh_token,
                                   expires=jwt.refresh_token_expires(now), path='/api/user/refresh'))


@ns.route('/me',
          method=Method.GET,
          access=AccessType.Fresh,
          returns=ns.route.returns('user', 'private'))
async def me(current_user, **_):
    return response.ok('OK', current_user)


@ns.route('/info',
          method=Method.GET,
          access=AccessType.Required,
          returns=ns.route.returns('user', 'private'))
async def info(current_user, **_):
    return response.ok('OK', current_user)


@ns.route('/login',
          method=Method.POST,
          returns=ns.route.returns('user', 'private'),
          expects=ns.route.expects('user', 'login'))
async def login(payload, **_):
    username = payload['username']
    password = payload['password']
    try:
        user = await user_service.get_by_username(username)
    except EntityNotFoundError:
        return response.unauthorized('user.login.wrong_credentials')
    if user is not None:
        if await user_service.check_password(user, password):
            resp = response.ok('user.login.success', user)
            _create_tokens(resp, user, set_access=True, set_refresh=True, fresh=True)
            return resp
    return response.unauthorized('user.login.wrong_credentials')


@ns.route('/logout',
          method=Method.POST)
async def logout(**_):
    resp = response.ok('user.logout.success')
    resp.cookies.append(Cookie('access_token', None, delete=True, path='/api'))
    resp.cookies.append(Cookie('refresh_token', None, delete=True, path='/api/user/refresh'))
    return resp


@ns.route('/token/refresh',
          method=Method.POST,
          access=AccessType.Refresh)
async def refresh(current_user, **_):
    resp = response.ok('user.token.refreshed')
    _create_tokens(resp, current_user, set_access=True, set_refresh=False, fresh=False)
    return resp


@ns.route('/register',
          method=Method.POST,
          returns=ns.route.returns('user', 'private'),
          expects=ns.route.expects('user', 'register'))
async def register(payload, **_):
    if env.init.get('ADMIN_REGISTER_ONLY', True):
        raise BadRequestError('global.register.admin_only')
    user = await user_service.create(payload)
    resp = response.created('user.registered', user)
    _create_tokens(resp, user, set_access=True, set_refresh=True, fresh=True)
    return resp


@ns.route('/register/admin',
          method=Method.POST,
          roles=['admin'],
          returns=ns.route.returns('user', 'private'),
          expects=ns.route.expects('user', 'admin_register'))
async def admin_register(payload, **_):
    send_mail = payload.pop('send_mail')
    payload['password'] = ''.join(random.choices(string.ascii_lowercase, k=32))
    user = await user_service.create(payload)
    if send_mail:
        await mail.sender.send(payload['email'], 'Welcome!', 'Welcome to Bolinette!')
    return response.created('user.registered', user)


@ns.route('/me',
          method=Method.PATCH,
          access=AccessType.Fresh,
          returns=ns.route.returns('user', 'private'),
          expects=ns.route.expects('user', 'register', patch=True))
async def update_user(payload, current_user, **_):
    user = await user_service.patch(current_user, payload)
    resp = response.ok('user.updated', user)
    _create_tokens(resp, user, set_access=True, set_refresh=True, fresh=True)
    return resp


ns.defaults.get_all('private', access=AccessType.Required, roles=['admin'])

ns.defaults.get_first_by('username', returns='private', access=AccessType.Required, roles=['admin'])


@ns.route('/{username}/roles',
          method=Method.POST,
          access=AccessType.Required,
          roles=['admin'],
          expects=ns.route.expects('role'),
          returns=ns.route.returns('user', 'private'))
async def add_user_role(match, payload, **_):
    user = await user_service.get_by_username(match['username'])
    role = await role_service.get_by_name(payload['name'])
    await user_service.add_role(user, role)
    return response.created(f'user.roles.added:{user.username}:{role.name}', user)


@ns.route('/{username}/roles/{role}',
          method=Method.DELETE,
          access=AccessType.Required,
          roles=['admin'],
          returns=ns.route.returns('user', 'private'))
async def delete_user_role(match, current_user, **_):
    user = await user_service.get_by_username(match['username'])
    role = await role_service.get_by_name(match['role'])
    await user_service.remove_role(current_user, user, role)
    return response.ok(f'user.roles.removed:{user.username}:{role.name}', user)


@ns.route('/picture',
          method=Method.POST,
          access=AccessType.Required,
          returns=ns.route.returns('user', 'private'))
async def upload_profile_picture(current_user, payload, **_):
    picture = payload['file']
    user = await user_service.save_profile_picture(current_user, picture)
    return response.ok(f'user.picture.uploaded', user)



from aiohttp import web
from aiohttp.web_request import Request

from bolinette import transaction, mapping
from bolinette.exceptions import AbortRequestException, ForbiddenError
from bolinette.routing import deserialize, serialize, AccessType
from bolinette.services import user_service
from bolinette.utils import Pagination


class Route:
    def __init__(self, *, func, base_url, path, method, access, expects, returns, roles):
        self.func = func
        self.base_url = base_url
        self.path = path
        self.method = method
        self.access = access
        self.expects = expects
        self.returns = returns
        self.roles = roles
        if self.roles and len(self.roles) and not self.access:
            self.access = AccessType.Required

    async def handler(self, request: Request):
        route_params = {
            'payload': await deserialize(request),
            'match': {},
            'query': {},
            'current_user': None,
        }
        for key in request.match_info:
            route_params['match'][key] = request.match_info[key]
        for key in request.query:
            route_params['query'][key] = request.query[key]
        try:
            with transaction:
                if self.access is not None:
                    route_params['current_user'] = await user_service.get_by_username(self.access.check(request))

                if self.roles and len(self.roles):
                    current_user = route_params['current_user']
                    user_roles = set(map(lambda r: r.name, current_user.roles))
                    if 'root' not in user_roles and not len(user_roles.intersection(set(self.roles))):
                        raise ForbiddenError(f'user.forbidden:{",".join(self.roles)}')

                if self.expects is not None:
                    def_key = f'{self.expects.model}.{self.expects.key}'
                    exp_def = mapping.get_payload(def_key)
                    route_params['payload'] = mapping.validate_payload(
                        exp_def, route_params['payload'], self.expects.patch)
                    mapping.link_foreign_entities(exp_def, route_params['payload'])

                resp = await self.func(**route_params)

            if isinstance(resp, web.Response):
                return resp

            content = resp.content

            if content.get('data') is not None and isinstance(content['data'], Pagination):
                content['pagination'] = {
                    'page': content['data'].page,
                    'per_page': content['data'].per_page,
                    'total': content['data'].total,
                }
                content['data'] = content['data'].items

            if self.returns is not None:
                def_key = f'{self.returns.model}.{self.returns.key}'
                ret_def = mapping.get_response(def_key)
                if content.get('data') is not None:
                    content['data'] = mapping.marshall(ret_def, content['data'],
                                                       self.returns.skip_none, self.returns.as_list)

            serialized, mime = serialize(content, 'application/json')

            web_response = web.Response(text=serialized, status=resp.code, content_type=mime)
            for cookie in resp.cookies:
                if not cookie.delete:
                    web_response.set_cookie(cookie.name, cookie.value,
                                            expires=cookie.expires.strftime('%a, %d %b %Y %H:%M:%S GMT'),
                                            path=cookie.path, httponly=cookie.http_only)
                else:
                    web_response.del_cookie(cookie.name, path=cookie.path)

            return web_response
        except AbortRequestException as ex:
            serialized, mime = serialize(ex.response.content, 'application/json')
            web_response = web.Response(text=serialized, status=ex.response.code, content_type=mime)
            return web_response

    def __str__(self):
        return f'<Route {self.method.http_verb} {self.base_url}{self.path} >'



from bolinette import response
from bolinette.routing import Method


class Defaults:
    def __init__(self, namespace):
        self.model = namespace.model
        self.service = namespace.service
        self.route = namespace.route

    def get_all(self, returns='default', *, access=None, roles=None):
        async def route(*, query, **_):
            pagination = None
            order_by = []
            if 'page' in query or 'per_page' in query:
                pagination = {
                    'page': int(query.get('page', 0)),
                    'per_page': int(query.get('per_page', 20))
                }
            if 'order_by' in query:
                columns = query['order_by'].split(',')
                for column in columns:
                    order_args = column.split(':')
                    col_name = order_args[0]
                    order_way = order_args[1] if len(order_args) > 1 else 'asc'
                    order_by.append((col_name, order_way == 'asc'))
            return response.ok('OK', await self.service.get_all(pagination, order_by))

        self.route('',
                   method=Method.GET,
                   returns=self.route.returns(self.model, returns, as_list=True),
                   access=access, roles=roles)(route)

    def get_one(self, returns='default', *, access=None, roles=None):
        async def route(*, match, **_):
            return response.ok('OK', await self.service.get(match.get('id')))

        self.route('/{id}',
                   method=Method.GET,
                   returns=self.route.returns(self.model, returns),
                   access=access, roles=roles)(route)

    def get_first_by(self, key, returns='default', *, access=None, roles=None):
        async def route(*, match, **_):
            return response.ok('OK', await self.service.get_first_by(key, match.get('id')))

        self.route('/{id}',
                   method=Method.GET,
                   returns=self.route.returns(self.model, returns),
                   access=access, roles=roles)(route)

    def create(self, returns='default', expects='default', *, access=None, roles=None):
        async def route(*, payload, **kwargs):
            return response.created(f'{self.model}.created', await self.service.create(payload, **kwargs))

        self.route('',
                   method=Method.POST,
                   returns=self.route.returns(self.model, returns),
                   expects=self.route.expects(self.model, expects),
                   access=access, roles=roles)(route)

    def update(self, returns='default', expects='default', *, access=None, roles=None):
        async def route(*, match, payload, **kwargs):
            entity = await self.service.get(match.get('id'))
            return response.ok(f'{self.model}.updated', await self.service.update(entity, payload, **kwargs))

        self.route('/{id}',
                   method=Method.PUT,
                   returns=self.route.returns(self.model, returns),
                   expects=self.route.expects(self.model, expects),
                   access=access, roles=roles)(route)

    def patch(self, returns='default', expects='default', *, access=None, roles=None):
        async def route(*, match, payload, **kwargs):
            entity = await self.service.get(match.get('id'))
            return response.ok(f'{self.model}.updated', await self.service.patch(entity, payload, **kwargs))

        self.route('/{id}',
                   method=Method.PATCH,
                   returns=self.route.returns(self.model, returns),
                   expects=self.route.expects(self.model, expects, patch=True),
                   access=access, roles=roles)(route)

    def delete(self, returns='default', *, access=None, roles=None):
        async def route(*, match, **kwargs):
            entity = await self.service.get(match.get('id'))
            return response.ok(f'{self.model}.deleted', await self.service.delete(entity, **kwargs))

        self.route('/{id}',
                   method=Method.DELETE,
                   returns=self.route.returns(self.model, returns),
                   access=access, roles=roles)(route)



from aiohttp import web

from bolinette.routing import Namespace, Method, AccessType
from bolinette.services import file_service

ns = Namespace('/file', file_service)

ns.defaults.get_first_by('key')


@ns.route('/{key}/download',
          method=Method.GET,
          access=AccessType.Fresh)
async def download_file(match, **_):
    file = await file_service.get_first_by('key', match['key'])
    headers = {
        'Content-disposition': f'attachment; filename={file.name}'
    }
    return web.Response(body=file_service.file_sender(file.key),
                        headers=headers, content_type=file.mime)



from functools import wraps
import pytest

from bolinette import Bolinette, db
from bolinette.routing import web
from bolinette.testing import TestClient


@pytest.fixture
def client(loop, aiohttp_client):
    Bolinette(profile='test',
              overrides={'DBMS': 'SQLITE', 'SECRET_KEY': 'super secret'})
    client = loop.run_until_complete(aiohttp_client(web.app))
    return TestClient(client)


def bolitest(*, before=None, after=None):
    def wrapper(func):
        @wraps(func)
        async def inner(*args, **kwargs):
            await db.engine.drop_all()
            await db.engine.create_all()
            if before is not None and callable(before):
                before()
            db.engine.session.commit()
            await func(*args, **kwargs)
            if after is not None and callable(after):
                after()
            await db.engine.drop_all()
        return inner
    return wrapper



from aiohttp import web as aio_web
import aiohttp_cors


class Web:
    def __init__(self):
        self.app = None
        self.cors = None

    def init_app(self):
        self.app = aio_web.Application()
        self.cors = aiohttp_cors.setup(self.app)


web = Web()



from aiohttp import web as aio_web

from bolinette import env
from bolinette.routing import web
from bolinette.commands import command


@command('run_server')
def run_server():
    aio_web.run_app(web.app, port=env['PORT'])



from bolinette import bcrypt
from bolinette.models import User, File
from bolinette.services import BaseService
from bolinette.exceptions import ForbiddenError, BadRequestError


class UserService(BaseService):
    @staticmethod
    def encrypt_password(password):
        return bcrypt.hash_password(password)

    def __init__(self):
        super().__init__(User)

    async def get_by_username(self, username):
        return await self.get_first_by('username', username)

    async def get_by_email(self, email):
        return await self.get_first_by('email', email)

    async def check_password(self, user, password):
        return bcrypt.check_password(user.password, password)

    async def _check_params(self, params):
        if params.get('password'):
            params['password'] = UserService.encrypt_password(params['password'])
        if params.get('timezone') and not await self.service('tz').is_valid(params['timezone']):
            raise BadRequestError(f'timezone.invalid:{params["timezone"]}')

    async def create(self, params, **_):
        await self._check_params(params)
        return await super().create(params)

    async def update(self, entity, params, **_):
        await self._check_params(params)
        return await super().update(entity, params)

    async def patch(self, entity, params, **_):
        await self._check_params(params)
        return await super().patch(entity, params)

    async def add_role(self, user, role):
        if role.name == 'root':
            raise ForbiddenError('role.root.forbidden')
        if role in user.roles:
            raise BadRequestError(f'user.roles.exists:{user.username}:{role.name}')
        user.roles.append(role)

    async def remove_role(self, current_user, user, role):
        if role.name == 'root':
            raise ForbiddenError('role.root.forbidden')
        if (current_user.username == user.username
                and role.name == 'admin'
                and not current_user.has_role('root')):
            raise ForbiddenError('role.admin.no_self_demotion')
        if role not in user.roles:
            raise BadRequestError(f'user.roles.not_found:{user.username}:{role.name}')
        user.roles.remove(role)

    async def save_profile_picture(self, user, request_file):
        old_picture = user.profile_picture
        user.profile_picture = await self.service(File).save_file(request_file)
        if old_picture:
            await self.service(File).delete(old_picture)
        return user


user_service = UserService()



from datetime import datetime

from bolinette.services import BaseService


class HistorizedService(BaseService):
    def __init__(self, model):
        super().__init__(model)

    async def create(self, params, *, current_user=None, **_):
        if current_user:
            now = datetime.utcnow()
            params['created_on'] = now
            params['created_by_id'] = current_user.id
            params['updated_on'] = now
            params['updated_by_id'] = current_user.id
        return await super().create(params, **_)

    async def update(self, entity, params, *, current_user=None, **_):
        if current_user:
            now = datetime.utcnow()
            params['created_on'] = entity.created_on
            params['created_by_id'] = entity.created_by_id
            params['updated_on'] = now
            params['updated_by_id'] = current_user.id
        return await super().update(entity, params, **_)

    async def patch(self, entity, params, *, current_user=None, **_):
        if current_user:
            now = datetime.utcnow()
            params['updated_on'] = now
            params['updated_by_id'] = current_user.id
        return await super().patch(entity, params, **_)



from enum import unique, Enum, auto

from bolinette import jwt


@unique
class AccessType(Enum):
    All = auto()
    Optional = auto()
    Required = auto()
    Fresh = auto()
    Refresh = auto()

    def check(self, request):
        return _functions[self.value](request)


_functions = {
    AccessType.All.value: lambda _: None,
    AccessType.Optional.value: lambda request: jwt.verify(request, optional=True),
    AccessType.Required.value: lambda request: jwt.verify(request),
    AccessType.Fresh.value: lambda request: jwt.verify(request, fresh=True),
    AccessType.Refresh.value: lambda request: jwt.verify(request)
}



from sqlalchemy import create_engine

from bolinette import env, db
from bolinette.utils import logger


class Engine:
    def __init__(self):
        self.engine = None
        self.session = None
        self.seeders = []

    def init_app(self):
        self.engine = create_engine(self._create_uri(), echo=False)
        db.defs.Session.configure(bind=self.engine)
        self.session = db.defs.Session()

    def seeder(self, func):
        self.seeders.append(func)
        return func

    async def create_all(self):
        db.defs.model.metadata.create_all(self.engine)

    async def drop_all(self):
        db.defs.model.metadata.drop_all(self.engine)

    async def run_seeders(self):
        for func in self.seeders:
            await func()

    def _create_uri(self):
        dbms = env.get('DBMS', 'SQLITE').lower()
        if dbms == 'sqlite':
            return 'sqlite:///' + env.instance_path(env.get('SQLITE_FILE', f'{env["PROFILE"]}.db'))
        if dbms == 'memory':
            return 'sqlite://'
        if dbms == 'postgresql':
            return 'postgresql://' + env['DB_URL']
        logger.error(f'Unknown database system "{dbms}"')
        exit(1)


engine = Engine()



from sqlalchemy import Table, Column, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship, backref

from bolinette import db


class DatabaseDefs:
    def __init__(self):
        self.Session = sessionmaker()
        self.model = declarative_base()
        self.table = Table
        self.relationship = relationship
        self.backref = backref

    def column(self, col_type: db.TypeClasses.DataType, col_rel: db.TypeClasses.ForeignKey = None, *,
               name=None, primary_key=False, nullable=True, unique=False):
        db_type = col_type.sql_alchemy_type
        db_rel = None
        if isinstance(col_rel, db.TypeClasses.ForeignKey):
            db_rel = ForeignKey(f'{col_rel.model}.{col_rel.key}')
        if name:
            return Column(name, db_type, db_rel, primary_key=primary_key, nullable=nullable, unique=unique,
                          comment=str(col_type))
        return Column(db_type, db_rel, primary_key=primary_key, nullable=nullable, unique=unique,
                      comment=str(col_type))


defs = DatabaseDefs()



class MappingObject:
    pass


class Field(MappingObject):
    def __init__(self, field_type, *, key=None, name=None,
                 default=None, required=False, function=None, formatting=None):
        self.type = field_type
        self.key = key
        self.name = name
        self.required = required
        self.default = default
        self.function = function
        self.formatting = formatting
        if not self.name and self.key:
            self.name = self.key

    def __repr__(self):
        return f'<MappingField {self.key}:{self.type} -> {self.name}>'


class List(MappingObject):
    def __init__(self, name, element):
        self.name = name
        self.element = element

    def __repr__(self):
        return f'<MappingList {self.name}:[{repr(self.element)}]>'


class Definition(MappingObject):
    def __init__(self, name, model, key='default'):
        self.fields = []
        self.name = name
        self.key = f'{model}.{key}'

    def __repr__(self):
        return f'<MappingModel {self.key}>'



from bolinette.mapping.objects import Field, List, Definition
from bolinette.mapping.definitions import (
    register, marshall, get_payload, get_response, get_model, link_foreign_entities
)
from bolinette.mapping.functions import map_model, validate_model, validate_payload



import sys


class Console:
    def print(self, text, *, sep=' ', end='\n'):
        print(text, sep=sep, end=end)

    def error(self, text, *, sep=' ', end='\n'):
        print(text, file=sys.stderr, sep=sep, end=end)


console = Console()



import traceback

from sqlalchemy.exc import SQLAlchemyError

from bolinette import response, db
from bolinette.exceptions import APIError, AbortRequestException
from bolinette.utils import logger


class Transaction:
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            db.engine.session.rollback()
            if issubclass(exc_type, APIError):
                res = exc_val.response
            else:
                logger.error(str(exc_val))
                res = response.internal_server_error(str(exc_val))
                traceback.print_tb(exc_tb)
            raise AbortRequestException(res)
        else:
            try:
                db.engine.session.commit()
            except SQLAlchemyError as err:
                db.engine.session.rollback()
                res = response.internal_server_error(f'global.internal_error:{err}')
                raise AbortRequestException(res)


transaction = Transaction()



from bolinette import response
from bolinette.routing import Namespace, Method
from bolinette.services import timezone_service

ns = Namespace('/tz', timezone_service)


@ns.route('', method=Method.GET)
async def all_timezones(**_):
    return response.ok('OK', await timezone_service.get_all())



from sqlalchemy import desc

from bolinette import mapping, db
from bolinette.exceptions import EntityNotFoundError
from bolinette.utils import Pagination

_services = {}


class BaseService:
    def __init__(self, model):
        self.model = model
        self.name = model.__tablename__.lower()
        _services[model] = self

    def service(self, model):
        return _services.get(model)

    async def get(self, identifier, **_):
        entity = db.engine.session.query(self.model).get(identifier)
        if entity is None:
            raise EntityNotFoundError(model=self.name, key='id', value=identifier)
        return entity

    async def get_by(self, key, value, **_):
        return db.engine.session.query(self.model).filter_by(**{key: value}).all()

    async def get_first_by(self, key, value, **_):
        entity = db.engine.session.query(self.model).filter_by(**{key: value}).first()
        if entity is None:
            raise EntityNotFoundError(model=self.name, key=key, value=value)
        return entity

    async def get_by_criteria(self, criteria, **_):
        return db.engine.session.query(self.model).filter(criteria).all()

    async def get_all(self, pagination=None, order_by=None, **_):
        if order_by is None:
            order_by = []
        query = db.engine.session.query(self.model)
        if len(order_by) > 0:
            query = await BaseService._build_order_by(self.model, query, order_by)
        if pagination is not None:
            return await BaseService._paginate(query, pagination)
        return query.all()

    async def create(self, params, **_):
        params = mapping.validate_model(self.model, params)
        entity = self.model(**params)
        db.engine.session.add(entity)
        return entity

    async def update(self, entity, params, **_):
        mapping.map_model(self.model, entity, params)
        return entity

    async def patch(self, entity, params, **_):
        mapping.map_model(self.model, entity, params, patch=True)
        return entity

    async def delete(self, entity, **_):
        db.engine.session.delete(entity)
        return entity

    @staticmethod
    async def _build_order_by(model, query, params):
        order_by_query = []
        for col_name, way in params:
            if hasattr(model, col_name):
                column = getattr(model, col_name)
                if way:
                    order_by_query.append(column)
                else:
                    order_by_query.append(desc(column))
        return query.order_by(*order_by_query)

    @staticmethod
    async def _paginate(query, pagination):
        page = pagination['page']
        per_page = pagination['per_page']
        total = query.count()
        items = query.offset(page * per_page).limit(per_page).all()
        return Pagination(items, page, per_page, total)


class SimpleService:
    def __init__(self, name):
        self.name = name
        _services[name] = self



import pytz

from bolinette.services import SimpleService


class TimezoneService(SimpleService):
    def __init__(self):
        super().__init__('tz')

    async def get_all(self):
        return [tz for tz in pytz.all_timezones]

    async def is_valid(self, key):
        return key in pytz.all_timezones


timezone_service = TimezoneService()



from bolinette import mapping, db


class Role(db.defs.model):
    __tablename__ = 'role'

    id = db.defs.column(db.types.integer, primary_key=True)
    name = db.defs.column(db.types.string, unique=True, nullable=False)

    def __repr__(self):
        return f'<Role {self.name}>'

    @staticmethod
    def payloads():
        yield [
            mapping.Field(db.types.string, key='name', required=True)
        ]

    @staticmethod
    def responses():
        yield [
            mapping.Field(db.types.string, key='name')
        ]
        yield 'complete', [
            mapping.Field(db.types.string, key='name'),
            mapping.List('users', mapping.Definition('user', 'user'))
        ]


mapping.register(Role)



from datetime import datetime

from bolinette import db
from bolinette.db import TypeClasses
from bolinette.exceptions import ParamConflictError, ParamMissingError


def validate_model(model, params, **kwargs):
    excluding = kwargs.get('excluding', [])
    errors = []
    valid = {}
    for column in model.__table__.columns:
        key = column.key
        if column.primary_key or key in excluding:
            continue
        value = params.get(key, None)
        if column.unique and value is not None:
            criteria = getattr(model, key) == value
            if db.engine.session.query(model).filter(criteria).first() is not None:
                errors.append((key, value))
        valid[key] = value
    if len(errors) > 0:
        raise ParamConflictError(params=errors)
    return valid


def validate_payload(definition, params, patch=False):
    errors = []
    valid = {}
    for field in definition.fields:
        if patch and field.name not in params:
            continue
        value = params.get(field.name, None)
        if value and field.type.of_type(TypeClasses.Date):
            value = datetime.strptime(value, '%Y-%m-%dT%H:%M:%S')
        if not value or not len(str(value)):
            if field.required:
                errors.append(field.name)
            else:
                value = field.default
        valid[field.name] = value
    if len(errors) > 0:
        raise ParamMissingError(params=errors)
    return valid


def map_model(model, entity, params, patch=False):
    errors = []
    for column in model.__table__.columns:
        key = column.key
        if column.primary_key or (key not in params and patch):
            continue
        original = getattr(entity, key)
        new = params.get(key, None)
        if original == new:
            continue
        if column.unique and new is not None:
            if db.engine.session.query(model).filter(getattr(model, key) == new).first() is not None:
                errors.append((key, new))
        setattr(entity, key, new)
    if len(errors) > 0:
        raise ParamConflictError(params=errors)



import sqlalchemy


class DataType:
    def __init__(self, sql_alchemy_type):
        self.sql_alchemy_type = sql_alchemy_type

    def of_type(self, cls):
        return type(self) is cls

    def __repr__(self):
        return self.__class__.__name__

    def __str__(self):
        return self.__class__.__name__.lower()


class String(DataType):
    def __init__(self):
        super().__init__(sqlalchemy.String)


class Integer(DataType):
    def __init__(self):
        super().__init__(sqlalchemy.Integer)


class Float(DataType):
    def __init__(self):
        super().__init__(sqlalchemy.Float)


class Boolean(DataType):
    def __init__(self):
        super().__init__(sqlalchemy.Boolean)


class Date(DataType):
    def __init__(self):
        super().__init__(sqlalchemy.DateTime)


class Email(DataType):
    def __init__(self):
        super().__init__(sqlalchemy.String)


class Password(DataType):
    def __init__(self):
        super().__init__(sqlalchemy.String)


class ForeignKey(DataType):
    def __init__(self, model, key):
        super().__init__(sqlalchemy.ForeignKey)
        self.model = model
        self.key = key


class TypeClasses:
    DataType = DataType
    String = String
    Integer = Integer
    Float = Float
    Boolean = Boolean
    Date = Date
    Email = Email
    Password = Password
    ForeignKey = ForeignKey


class Types:
    @property
    def string(self):
        return String()

    @property
    def integer(self):
        return Integer()

    @property
    def float(self):
        return Float()

    @property
    def boolean(self):
        return Boolean()

    @property
    def date(self):
        return Date()

    @property
    def email(self):
        return Email()

    @property
    def password(self):
        return Password()

    def foreign_key(self, model, key):
        return ForeignKey(model, key)


types = Types()



from bolinette.environment import env
from bolinette.response import response, Cookie
from bolinette.jwt import jwt
from bolinette.transaction import transaction
from bolinette.bcrypt import bcrypt
from bolinette.commands import command
import bolinette.controllers
from bolinette.bolinette import Bolinette



from bolinette.controllers.file import ns as file_controller
from bolinette.controllers.user import ns as user_controller
from bolinette.controllers.role import ns as role_namespace
from bolinette.controllers.timezone import ns as tz_namespace



from setuptools import setup, find_packages

from bolinette import env
from bolinette.utils import fs


def project_packages(module):
    return [m for m in find_packages() if m.startswith(module)]


setup(
    name='Bolinette',
    packages=project_packages('bolinette'),
    include_package_data=True,
    version=fs.read_version(env.root_path()),
    license='MIT',
    description='Bolinette, a web framework built on top of Flask',
    author='Pierre Chat',
    author_email='pierrechat89@hotmail.fr',
    url='https://github.com/TheCaptainCat/bolinette',
    keywords=['Flask', 'Bolinette', 'Web', 'Framework'],
    install_requires=[
        'aiohttp==3.6.2',
        'aiohttp-cors==0.7.0',
        'bcrypt==3.1.7',
        'pytz==2019.3',
        'PyJWT==1.7.1',
        'pytest==5.4.1',
        'pytest-aiohttp==0.3.0',
        'PyYAML==5.3.1',
        'requests==2.23.0',
        'SQLAlchemy==1.3.15',
        'twine==3.1.1'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
    ],
    setup_requires=[
        'wheel'
    ]
)



from bolinette import response


class APIError(Exception):
    def __init__(self, name, function, messages):
        super().__init__(name)
        self.function = function
        self.messages = messages

    def __str__(self):
        return ", ".join(self.messages)

    @property
    def response(self):
        return self.function(self.messages)


class NotFoundError(APIError):
    def __init__(self, messages, *, name=None):
        super().__init__(name or type(self).__name__,
                         response.not_found, messages)


class ConflictError(APIError):
    def __init__(self, messages, *, name=None):
        super().__init__(name or type(self).__name__,
                         response.conflict, messages)


class BadRequestError(APIError):
    def __init__(self, messages, *, name=None):
        super().__init__(name or type(self).__name__,
                         response.bad_request, messages)


class ForbiddenError(APIError):
    def __init__(self, messages, *, name=None):
        super().__init__(name or type(self).__name__,
                         response.forbidden, messages)


class UnauthorizedError(APIError):
    def __init__(self, messages, *, name=None):
        super().__init__(name or type(self).__name__,
                         response.unauthorized, messages)


class EntityNotFoundError(NotFoundError):
    def __init__(self, **kwargs):
        params = kwargs.get('params', None)
        model = kwargs.get('model', None)
        key = kwargs.get('key', None)
        value = kwargs.get('value', None)
        if params is None:
            params = [(model, key, value)]
        messages = [f'{m}.not_found:{k}:{v}' for m, k, v in params]
        super().__init__(messages, name='EntityNotFoundError')


class ParamMissingError(BadRequestError):
    def __init__(self, **kwargs):
        params = kwargs.get('params', None)
        key = kwargs.get('key', None)
        if params is None:
            params = [key]
        messages = [f'param.required:{k}' for k in params]
        super().__init__(messages, name='ParamMissingError')


class ParamConflictError(ConflictError):
    def __init__(self, **kwargs):
        params = kwargs.get('params', None)
        key = kwargs.get('key', None)
        value = kwargs.get('value', None)
        if params is None:
            params = [(key, value)]
        messages = [f'param.conflict:{k}:{v}' for k, v in params]
        super().__init__(messages, name='ParamConflictError')


class AbortRequestException(Exception):
    def __init__(self, resp):
        self.response = resp



from bolinette import mapping, env, db
from bolinette.models import Role, File

users_roles = db.defs.table(
    'users_roles',
    db.defs.model.metadata,
    db.defs.column(db.types.integer, db.types.foreign_key('user', 'id'), name='u_id', primary_key=True),
    db.defs.column(db.types.integer, db.types.foreign_key('role', 'id'), name='r_id', primary_key=True))


class User(db.defs.model):
    __tablename__ = 'user'

    id = db.defs.column(db.types.integer, primary_key=True)
    username = db.defs.column(db.types.string, unique=True, nullable=False)
    password = db.defs.column(db.types.string, nullable=False)
    email = db.defs.column(db.types.string, unique=env.init['USER_EMAIL_REQUIRED'],
                           nullable=(not env.init['USER_EMAIL_REQUIRED']))

    roles = db.defs.relationship(Role, secondary=users_roles, lazy='subquery',
                                 backref=db.defs.backref('user', lazy=True))

    picture_id = db.defs.column(db.types.integer, db.types.foreign_key('file', 'id'))
    profile_picture = db.defs.relationship(File, foreign_keys=picture_id, lazy=False)

    timezone = db.defs.column(db.types.string)

    def has_role(self, role):
        return any(filter(lambda r: r.name == role, self.roles))

    @staticmethod
    def payloads():
        yield 'register', [
            mapping.Field(db.types.string, key='username', required=True),
            mapping.Field(db.types.email, key='email', required=True),
            mapping.Field(db.types.password, key='password', required=True),
            mapping.Field(db.types.string, key='timezone')
        ]
        yield 'admin_register', [
            mapping.Field(db.types.string, key='username', required=True),
            mapping.Field(db.types.email, key='email', required=True),
            mapping.Field(db.types.boolean, key='send_mail', required=True)
        ]
        yield 'login', [
            mapping.Field(db.types.string, key='username', required=True),
            mapping.Field(db.types.password, key='password', required=True)
        ]

    @staticmethod
    def responses():
        default = [
            mapping.Field(db.types.string, key='username'),
            mapping.Definition('profile_picture', 'file', 'minimal')
        ]
        yield default
        yield 'private', default + [
            mapping.Field(db.types.email, key='email'),
            mapping.List('roles', mapping.Definition('role', 'role')),
            mapping.Field(db.types.string, key='timezone')
        ]

    def __repr__(self):
        return f'<User {self.username}>'


mapping.register(User)



from bolinette.services.base import BaseService, SimpleService
from bolinette.services.historized import HistorizedService
from bolinette.services.user import user_service
from bolinette.services.role import role_service
from bolinette.services.file import file_service
from bolinette.services.timezone import timezone_service



from bolinette.routing.web import web
from bolinette.routing.resources import resources
from bolinette.routing.serializing import deserialize, serialize
from bolinette.routing.access import AccessType
from bolinette.routing.method import Method
from bolinette.routing.route import Route
from bolinette.routing.namespace import Namespace
